--[[
YAGRA - gear rating addon
Copyright 2010 Adirelle (adirelle@tagada-team.net)
All rights reserved.
--]]

local addonName, addon = ...
local L = addon.L

local databaseAddonName = addonName..'_Database'

local NormalizeName = addon.NormalizeName
local playerName, playerRealm, playerFaction = addon.playerName, addon.playerRealm, addon.playerFaction

local RECORD_SEPARATOR = '|'
local RECORD_VERSION = 4

local SERIALIZED_RACE = {
	Human = "HU",
	Dwarf = "DW",
	Gnome = "GN",
	Draenei = "DR",
	NightElf = "NE",
	Orc = "OR",
	Tauren = "TA",
	Troll = "TR",
	BloodElf = "BE",
	Scourge = "UN",
	Worgen = "WO",
	Goblin = "GO",
	Undead = "UN", -- should not occur
}
local UNSERIALIZED_RACE = {
	HU = "Human",
	DW = "Dwarf",
	GN = "Gnome",
	DR = "Draenei",
	NE = "NightElf",
	OR = "Orc",
	TA = "Tauren",
	TR = "Troll",
	BE = "BloodElf",
	UN = "Scourge",
	WO = "Worgen",
	GO = "Goblin",
}

local SERIALIZED_CLASS = {
	WARRIOR = "WA",
	DEATHKNIGHT = "DK",
	PALADIN = "PA",
	SHAMAN = "SH",
	HUNTER = "HU",
	DRUID = "DR",
	ROGUE = "RO",
	WARLOCK = "WK",
	PRIEST = "PR",
	MAGE = "MA",
}
local UNSERIALIZED_CLASS = {
	WA = "WARRIOR",
	DK = "DEATHKNIGHT",
	PA = "PALADIN",
	SH = "SHAMAN",
	HU = "HUNTER",
	DR = "DRUID",
	RO = "ROGUE",
	WK = "WARLOCK",
	PR = "PRIEST",
	MA = "MAGE",
}

local SERIALIZED_FACTION = {
	Horde = "H",
	Alliance = "A",
}
local UNSERIALIZED_FACTION = {
	H = "Horde",
	A = "Alliance",
}

-- Export (un)serialization maps
addon.SERIALIZED_RACE = SERIALIZED_RACE
addon.UNSERIALIZED_RACE = UNSERIALIZED_RACE
addon.SERIALIZED_CLASS = SERIALIZED_CLASS
addon.UNSERIALIZED_CLASS = UNSERIALIZED_CLASS
addon.SERIALIZED_FACTION = SERIALIZED_FACTION
addon.UNSERIALIZED_FACTION = UNSERIALIZED_FACTION

local tostring = tostring
local tonumber = tonumber
local rawset = rawset
local rawget = rawget
local unpack = unpack
local assert = assert
local strjoin = strjoin
local strsplit = strsplit

local function SerializeRecord(record)
	local str = strjoin(RECORD_SEPARATOR, tostringall(
		RECORD_VERSION,
		record.timestamp,
		record.score,
		record.average,
		record.level,
		SERIALIZED_CLASS[record.class],
		SERIALIZED_RACE[record.race],
		SERIALIZED_FACTION[record.faction],
		record.resilience,
		tostring(record.ratingScheme or 0),
		unpack(record, 1, 18)
	))
	if record.timestamp and record.score and record.average and record.level and record.resilience and SERIALIZED_CLASS[record.class or ""]
			and SERIALIZED_RACE[record.race or ""] and SERIALIZED_FACTION[record.faction or ""] then
		return str
	else
		geterrorhandler()("Invalid record for "..tostring(record.name)..": "..str:sub(1,50))
	end
end

local UnserializeRecord
do
	local handlers
	handlers = {
		-- Version 1
		function(t, timestamp, score, average, level, class, race, faction, ...)
			t.timestamp = tonumber(timestamp)
			t.score = tonumber(score)
			t.average = tonumber(average)
			t.level = tonumber(level)
			t.class = UNSERIALIZED_CLASS[class]
			t.race = UNSERIALIZED_RACE[race]
			t.faction = UNSERIALIZED_FACTION[faction]
			t.resilience = 0
			for i = 1, 18 do
				t[i] = select(i, ...)
			end
		end,
		-- Version 2
		function(t, timestamp, score, average, level, class, race, faction, resilience, ...)
			handlers[1](t, timestamp, score, average, level, class, race, faction, ...)
			t.resilience = tonumber(resilience)
		end,
		-- Version 3
		function(t, timestamp, score, average, level, class, race, faction, resilience, alternativeRating, ...)
			handlers[2](t, timestamp, score, average, level, class, race, faction, resilience, ...)
			t.alternativeRating = (alternativeRating == "A")
		end,
		-- Version 4
		function(t, timestamp, score, average, level, class, race, faction, resilience, ratingScheme, ...)
			handlers[3](t, timestamp, score, average, level, class, race, faction, resilience, nil, ...)
			t.alternativeRating = nil
			t.ratingScheme = tonumber(ratingScheme)
		end,	
	}
	local function FillRecord(t, version, ...)
		version = tonumber(version)
		local handler = handlers[version]
		assert(handler, "Unknown record version:"..tostring(version))
		handler(t, ...)
		if version < RECORD_VERSION then
			t.__dirty = true
		end
	end
	function UnserializeRecord(name, realm, data, t)
		if not t then
			t = {}
		end
		t.fullName, t.name, t.realm = NormalizeName(name, realm)
		FillRecord(t, strsplit(RECORD_SEPARATOR, data))
		return t
	end
end

local database

function addon:InitializeDatabase()
	-- Initialize the database if need be
	if not YAGRAPlayerDB then
		YAGRAPlayerDB = {}
	end
	database = YAGRAPlayerDB
	self.db.RegisterCallback(self, 'OnDatabaseShutdown')
end

function addon:OnDatabaseShutdown()
	if self.db.profile.autoPrune and time() - (database.__lastPruningTime or 0) > 86400 then
		self:PruneDatabase()
	else
		self:SaveDirtyRecords()
	end
end

function addon:CleanupDatabase(event)
	if event == 'PLAYER_UPDATE_RESTING' then
		addon:SaveDirtyRecords()
		addon:WipeCache()
	elseif not addon:HasDirtyRecords() then
		addon:WipeCache()
	end
end

function addon:IsRecordValid(record)
	return record and record.timestamp >= time() - 86400 * self.db.profile.retentionPeriod and record.level >= self.db.profile.minLevel
end

local dirty = 0

function addon:HasDirtyRecords()
	return dirty > 0
end

local recordMeta = {
	__newindex = function(self, key, value)
		assert(key, "Invalid record key: "..tostring(key))
		if value == self[key] then return end
		rawset(self, key, value)
		if not self.__dirty and tostring(key):sub(1,2) ~= '__' then
			dirty = dirty + 1
			rawset(self, '__dirty', true)
		end
	end
}

local records = setmetatable({}, {
	__index = function(self, realm)
		assert(type(realm) == "string", "Invalid realm name: "..tostring(realm))
		local realmRecords = setmetatable({}, {
			__index = function(self, name)
				assert(type(name) == "string", "Invalid character name :"..tostring(name))
				local data = database[realm] and database[realm][name]
				if data then
					local record = UnserializeRecord(name, realm, data)
					if addon:IsRecordValid(record) then
						setmetatable(record, recordMeta)
						rawset(self, name, record)
						return record
					else
						-- Invalid or outdated record
						database[name] = nil
					end
				end
			end,
		})
		rawset(self, realm, realmRecords)
		return realmRecords
	end
})

local function RealmIterator(t, realm)
	local records
	repeat
		realm, records = next(t, realm)
		if not realm then
			return
		end
	until realm and realm:sub(1,2) ~= "__" and type(records) == "table"
	return realm, records
end
local function IterateRealms(t)
	return RealmIterator, t
end

local statCache = {}

function addon:SaveDirtyRecords()
	if dirty == 0 then return end
	self:Debug('Saving dirty records')
	local count = 0
	for realm, realmRecords in IterateRealms(records) do
		for name, record in pairs(realmRecords) do
			if record.__dirty then
				if not database[realm] then database[realm] = {} end
				database[realm][name] = SerializeRecord(record)
				rawset(record, '__dirty', nil)
				count = count + 1
			end
		end
	end
	dirty = 0
	wipe(statCache)
	self:Debug(count, 'records saved')
end

function addon:WipeCache()
	self:Debug('Wiping working database')
	wipe(records)
	wipe(statCache)
end

function addon:PruneDatabase(dryRun)
	self:SaveDirtyRecords()
	local t = {}
	local count = 0
	for realm, realmRecords in IterateRealms(database) do
		for name, data in pairs(realmRecords) do
			local record = UnserializeRecord(name, realm, data, t)
			if not self:IsRecordValid(record) then
				if not dryRun then
					realmRecords[name] = nil
				end
				count = count + 1
			end
		end
		if not dryRun and not next(realmRecords) then
			database[realm] = nil
		end
	end
	if not dryRun then
		database.__lastPruningTime = time()
		self:Feedback(L['Removed %d |4record:records;.']:format(count))
		self:WipeCache()
	end
	return count
end

function addon:ClearDatabase()
	self:WipeCache()
	wipe(database)
end

local function RefreshRecord(record)
	if record and record.ratingScheme ~= addon:GetRatingSchemeID() then
		addon:UpdatePlayerScore(record, "Database", true)
	end
	return record
end

function addon:SearchDatabase(namePattern, realmPattern)
	self:SaveDirtyRecords()
	namePattern = (namePattern and namePattern ~= '*') and ('^'..namePattern:gsub('%*', '.-'):lower()..'$')
	realmPattern = (realmPattern and realmPattern ~= '*') and ('^'..realmPattern:gsub('%*', '.-'):lower()..'$')
	local realm, realmRecords, name
	return function()
		repeat
			if not realmRecords then
				repeat
					realm, realmRecords = RealmIterator(database, realm)
					if not realm then
						return
					end
				until not realmPattern or realm:lower():match(realmPattern)
			end
			name = next(realmRecords, name)
			if not name then
				realmRecords = nil
			end
		until name and (not namePattern or name:lower():match(namePattern))
		return name, realm, RefreshRecord(records[realm][name])
	end
end

function addon:GetPlayerRecord(name, realm)
	local _, name, realm = NormalizeName(name, realm)
	return RefreshRecord(records[realm] and records[realm][name])
end

function addon:CreatePlayerRecord(name, realm, record)
	local fullName, name, realm = NormalizeName(name, realm)
	assert(not self:GetPlayerRecord(name, realm), "Record for "..fullName.." already exists")
	if not record then record = {} end
	record.name = name
	record.realm = realm
	record.fullName = fullName
	record.timestamp = record.timestamp or time()
	if not record.__dirty then
		record.__dirty = true
		dirty = dirty + 1
	end
	setmetatable(record, recordMeta)
	records[realm][name] = record
	return record
end

function addon:GetStats(force)
	if not force and statCache.timestamp and GetTime() - statCache.timestamp < 5 then
		return statCache
	end
	wipe(statCache)
	statCache.timestamp = GetTime()
	statCache.memUsage = L['%dkB']:format(math.ceil(GetAddOnMemoryUsage(addonName)))
	if IsAddOnLoaded(databaseAddonName) then
		statCache.dbMemUsage = L['%dkB']:format(math.ceil(GetAddOnMemoryUsage(databaseAddonName)))
		local total = 0
		for realm, realmRecords in IterateRealms(database) do
			for name in pairs(realmRecords) do
				total = total + 1
			end
		end
		statCache.dbSize = total
	else
		statCache.dbMemUsage = L['not yet loaded']
		statCache.dbSize = L['not yet loaded']
	end
	local loaded  = 0
	dirty = 0
	for realm, realmRecords in IterateRealms(records) do
		for name, record in pairs(realmRecords) do
			loaded = loaded  + 1
			if record.__dirty then
				dirty = dirty + 1
			end
		end
	end
	statCache.loaded = loaded
	statCache.dirty = dirty
	return statCache
end

function addon:PrintDatabaseStats()
	local stats = self:GetStats(true)
	self:Feedback(L['Statistics'])
	self:Feedback(L['- main addon memory usage: %s']:format(stats.memUsage))
	self:Feedback(L['- database memory usage: %s']:format(stats.dbMemUsage))
	self:Feedback(L['- total number of player records: %d']:format(stats.dbSize))
	self:Feedback(L['- number of loaded records: %d (%d dirty)']:format(stats.loaded, stats.dirty))
end
