--[[
YAGRA - gear rating addon
Copyright 2010 Adirelle (adirelle@tagada-team.net)
All rights reserved.
--]]

local addonName, addon = ...
local L = addon.L

local handlerProto = {}
local handlerMeta = { __index = handlerProto }

function handlerProto:ResolvePath(info)
	local db = self.dbHolder.db.profile
	local path = info.arg or info[#info]
	if type(path) == "string" then
		return db, path, path
	elseif type(path) == "table" then
		local n = #path
		for i = 1, n-1 do
			db = db[path[i]]
		end
		return db, path[n], strjoin('.', unpack(path))
	end
end

function handlerProto:Get(info, ...)
	local db, key = self:ResolvePath(info)
	if info.type == "multiselect" then
		local subKey = ...
		return db[key] and db[key][subKey]
	else
		return db[key]
	end
end

function handlerProto:Set(info, value, ...)
	local db, key, path = self:ResolvePath(info)
	if info.type == 'multiselect' then
		local subKey, value = value, ...
		db[key][subKey] = value
		path = strjoin('.', path, subKey)
	else
		db[key] = value
	end
	self.dbHolder:Debug('ConfigSet', path, value)
	self.dbHolder:SendMessage('YAGRA_ConfigChanged', path, value)
end

function handlerProto:IsDisabled(info)
	if self.dbHolder == addon then
		return info.type ~= "group" and not addon.db.profile.enabled
	else
		return info.type ~= "group" and (not addon.db.profile.enabled or not addon.db.profile.modules[self.dbHolder.moduleName])
	end
end

function addon.GetOptionHandler(dbHolder)
	return setmetatable({dbHolder = dbHolder}, handlerMeta)
end

addon.modulePrototype.GetOptionHandler = addon.GetOptionHandler

function addon.modulePrototype:CreateEnableOption(t)
	if not t then
		t = {
			name = L[self.moduleName],
			desc = L["Enable " .. self.moduleName],
		}
	end
	t.type = 'toggle'
	t.handler = addon.options.handler
	t.arg = { 'modules', self.moduleName }
	t.disabled = "IsDisabled"
	return t
end

function addon:RefreshOptions()
	if not self.options then return end
	LibStub('AceConfigRegistry-3.0'):NotifyChange(addonName)
end

function addon:RegisterModuleOptions(module)
	if not self.options then return end
	if module.GetOptions then
		local name = module.moduleName
		local options, handler = module:GetOptions()
		if options.type == 'group' then
			options = { [name] = options }
			self.options.plugins[name] = options
		else
			self.options.args.options.plugins[name] = options
		end
		if handler then
			for key, option in pairs(options) do
				if option.type and not option.handler then
					option.handler = handler
				end
			end
		end
		self:Debug('Registered options of module', name)
	end
end

function addon.GetOptions()
	if not addon.options then
		addon.options = {
			name = 'YAGRA',
			type = 'group',
			childGroups = 'tab',
			handler = addon:GetOptionHandler(),
			get = "Get",
			set = "Set",
			disabled = "IsDisabled",
			plugins = {},
			args = {
				conf = {
					name = L["Configuration"],
					desc = L["Open the configuration panel."],
					type = 'execute',
					guiHidden = true,
					func = function() InterfaceOptionsFrame_OpenToCategory(addon.optionPanel) end,
					order = 1,
				},
				query = {
					name = L["Query database"],
					desc = L["Query character data from the database."],
					type = 'input',
					guiHidden = true,
					get = function() end,
					set = function(_, input) addon:ChatQuery(input) end,
					order = 2,
				},
				options = {
					name = L['Options'],
					order = 3,
					type = 'group',
					plugins = {},
					args = {
						enabled = {
							name = L['Enable YAGRA'],
							type = 'toggle',
							order = 2,
							disabled = false,
						},
						legacyQualityModifier = {
							name = L['Legacy quality modifiers'],
							desc = L['Apply quality modifiers to Cataclysm item scores, even though quality does not affect stat budget of Cataclysm items. Enable this to calculate scores compatible with legacy addons ; disable it to obtain more accurate scores.'],
							type = 'toggle',
							order = 3,							
						},
						alternativeRating = {
							name = L['Alternative weapon rating'],
							desc = L['Uses slightly different weightings for main hand weapons and off-hand weapons / holdables / shields. Scores calculated using that rating are marked with a trailing asterisk: *.'],
							type = 'toggle',
							order = 4,
						},
						minLevel = {
							name = L['Minimum level'],
							desc = L['Characters of level below this value are ignored ; existing records will be removed on database pruning.'],
							type = 'range',
							min = 1,
							max = MAX_PLAYER_LEVEL,
							order = 10,
						},
						crossRealm = {
							name = L['Cross realm'],
							desc = L['Check this to save the score of characters from other realms.'],
							type = 'toggle',
							order = 20,
						},
						crossFaction = {
							name = L['Cross faction'],
							desc = L['Check this to save the score of characters of the opposite faction.'],
							type = 'toggle',
							order = 30,
						},
						chatFeedback = {
							name = L['Chat output'],
							desc = L['Output a line in the chatframe whenever some player score is updated.'],
							type = 'toggle',
							order = 40,
						},
						_scoreHeader = {
							name = L['Score display'],
							type = 'header',
							order = 50,
						},
						scoreFormat = {
							name = L['Score format'],
							desc = L['Select how scores should be displayed'],
							type = 'select',
							order = 60,
							values = {
								score = L['Score'],
								score_itemlevel = strjoin(' / ', L['Score'], strlower(L['Average item level'])),
								itemlevel = L['Average item level'],
								tier_itemlevel = strjoin(' / ', L['Content tier'], strlower(L['Average item level'])),
								tier = L['Content tier'],
								score_tier = strjoin(' / ', L['Score'], strlower(L['Content tier']));
								score_tier_itemlevel = strjoin(' / ', L['Score'], strlower(L['Content tier']), strlower(L['Average item level'])),
							},
						},
						showDate = {
							name = L['Show player data age'],
							desc = L['Display how old the data about a player is.'],
							type = 'toggle',
							order = 70,
						},
						qualityColor = {
							name = L['Score coloring'],
							desc = L['Check this to color scores using a scheme based on item quality colors.'],
							type = 'toggle',
							order = 80,
						},
					},
				},
				database = {
					name = L['Database'],
					type = 'group',
					order = 4,
					args = {
						_statHeader = { name = L['Statistics'], type = 'header', order = 1010, cmdHidden = true, },
						memUsage = {
							name = function() return ("|cffffd100%s:|r %s"):format(L['Memory usage'], addon:GetStats().memUsage) end,
							type = 'description', order = 1020, width = 'double', cmdHidden = true,
						},
						dbMemUsage = {
							name = function() return ("|cffffd100%s:|r %s"):format(L['Database memory usage'] ,addon:GetStats().dbMemUsage) end,
							type = 'description', order = 1025, width = 'double', cmdHidden = true,
						},
						dbSize = {
							name = function() return ("|cffffd100%s:|r %d"):format(L['Number of player records'], addon:GetStats().dbSize) end,
							type = 'description', order = 1030, width = 'double', cmdHidden = true,
						},
						cacheSize = {
							name = function()
								local stats = addon:GetStats()
								return ("|cffffd100%s:|r %d ("..L['%d dirty']..")"):format(L['Number of loaded records'], stats.loaded, stats.dirty) end,
							type = 'description', order = 1035, width = 'double', cmdHidden = true,
						},
						stats = {
							name = L['Statistics'],
							desc = L['Print database statistics'],
							type = 'execute',
							func = function() addon:PrintDatabaseStats() end,
							guiHidden = true,
							order = 1100,
						},
						_settingsHeader = { name = L['Settings'], type = 'header', order = 1100, cmdHidden = true, },
						retention = {
							name = L['Retention period (days)'],
							desc = L['How many days should player records be kept in database.'],
							type = 'range',
							arg = 'retentionPeriod',
							min = 1,
							max = 60,
							step = 1,
							order = 1110,
						},
						prune = {
							name = L['Prune database'],
							desc = L['Prune outdated records from the database.'],
							confirm = function()
								return L["This will definitively remove %d record(s) from the database. Do you really want to do this ?"]:format(addon:PruneDatabase(true))
							end,
							type = 'execute',
							func = function()
								addon:PruneDatabase()
								addon:RefreshOptions()
							end,
							order = 1120,
						},
						autoPrune = {
							name = L['Auto-prune database'],
							desc = L['Automatically prune the database once a day.'],
							type = 'toggle',
							order = 1125,
						},
						clear = {
							name = L['Clear database'],
							desc = L['Destroy all saved player records.'],
							type = 'execute',
							confirm = true,
							confirmText = L["This will definitively destroy all saved data, with no way back. Do you really want to do this ?"],
							func = function()
								addon:ClearDatabase()
								addon:RefreshOptions()
							end,
							order = 1130,
						},
					},
				},
			},
		}
		for _, module in addon:IterateModules() do
			addon:RegisterModuleOptions(module)
		end
	end
	return addon.options
end
