--[[
YAGRA - gear rating addon
Copyright 2010 Adirelle (adirelle@tagada-team.net)
All rights reserved.
--]]

local addonName, addon = ...
local L = addon.L

LibStub('AceAddon-3.0'):NewAddon(addon, addonName, 'AceEvent-3.0', 'AceConsole-3.0')
--[===[@debug@
_G.YAGRA = addon
--@end-debug@]===]

--------------------------------------------------------------------------------
-- Debug stuff
--------------------------------------------------------------------------------

if AdiDebug then
	AdiDebug:Embed(addon, addonName)
else
	function addon.Debug() end
end

--------------------------------------------------------------------------------
-- Addon-wide constants
--------------------------------------------------------------------------------

addon.playerName, addon.playerRealm, addon.playerFaction = UnitName('player'), GetRealmName(), UnitFactionGroup("player")
local playerName, playerRealm, playerFaction = addon.playerName, addon.playerRealm, addon.playerFaction

--------------------------------------------------------------------------------
-- Addon initializing and enabling/disabling handlers
--------------------------------------------------------------------------------

function addon:OnInitialize()
	self.db = LibStub('AceDB-3.0'):New("YAGRADB", {
		profile = {
			enabled = true,
			modules = { ['*'] = true },
			retentionPeriod = 10,
			autoPrune = false,
			minLevel = 10,
			crossFaction = true,
			crossRealm = true,
			chatFeedback = true,
			qualityColor = true,
			showDate = true,
			scoreFormat = "score",
			legacyQualityModifier = true,
			alternativeRating = false,
		},
	}, true)
	LibStub('AceConfig-3.0'):RegisterOptionsTable(addonName, self.GetOptions)
	LibStub("AceConfigCmd-3.0"):CreateChatCommand("yagra", addonName)
	LibStub("AceConfigCmd-3.0"):CreateChatCommand("yg", addonName)
	self.optionPanel = LibStub('AceConfigDialog-3.0'):AddToBlizOptions(addonName, addonName)
	self:RegisterChatCommand("ygstat", "PrintDatabaseStats", true)
	self:RegisterChatCommand("ygquery", "ChatQuery", true)

	self:InitializeTiers()
	self:InitializeDatabase()

	self:SetEnabledState(self.db.profile.enabled)
	self:UpdateModuleStates('OnInitialize')

	-- Register YAGRA_ConfigChanged using a string as reference so it doesn't get unregistered on disable
	LibStub('AceEvent-3.0').RegisterMessage(addonName, 'YAGRA_ConfigChanged', self.YAGRA_ConfigChanged, self)
end

function addon:OnEnable()
	self:RegisterMessage('YAGRA_RecordUpdated')
	self:RegisterEvent('PLAYER_REGEN_ENABLED', 'UpdateModuleStates')
	self:RegisterEvent('PLAYER_REGEN_DISABLED', 'OnCombatStart')
	self:RegisterEvent('PLAYER_UPDATE_RESTING', 'CleanupDatabase')
	self:Debug('Enabled')
end

function addon:OnDisable()
	self:Debug('Disabled')
end

function addon:OnCombatStart(event)
	self:CleanupDatabase(event)
	self:UpdateModuleStates(event)
end

function addon:YAGRA_ConfigChanged(event, key, value)
	self:Debug('YAGRA_ConfigChanged', key, value)
	if key == "enabled" then
		if value then
			self:Enable()
		else
			self:Disable()
		end
	elseif key:match('^modules.') then
		self:UpdateModuleStates(event)
	else
		local ratingScheme = self:GetRatingSchemeID()
		if ratingScheme ~= self.ratingScheme then
			self:InitializeTiers()
			self:GetPlayerRecord(playerName, playerRealm)
			--local record = 
			--self:UpdatePlayerScore()
		end
	end
end

function addon:Feedback(...)
	print(strjoin(" ", '|cffff8800['..self.name..']|r', tostringall(...)))
end

--------------------------------------------------------------------------------
-- Some helpers
--------------------------------------------------------------------------------

local function UnitNameRealm(unit)
	local name, realm = UnitName(unit)
	if realm and realm ~= "" then
		return name, realm, strjoin('-', name, realm)
	else
		return name, playerRealm, name
	end
end

local function NormalizeName(name, realm)
	if name:match('-') then
		name, realm = strsplit('-', name)
	end
	if not realm or realm == "" or realm == playerRealm then
		return name, name, playerRealm
	else
		return strjoin('-', name, realm), name, realm
	end
end

addon.UnitNameRealm = UnitNameRealm
addon.NormalizeName = NormalizeName

function addon:AcceptPlayer(level, faction, realm)
	if not self.db.profile.crossRealm and (realm and realm ~= playerRealm) then
		return false, 'character from other realm'
	elseif not self.db.profile.crossFaction and faction ~= playerFaction then
		return false, 'character from other faction'
	elseif level < self.db.profile.minLevel then
		return false, 'character under level requirement'
	end
	return true
end

function addon:IsValidUnit(unit)
	if not unit then
		return false, 'invalid unit'
	elseif not UnitExists(unit)  then
		return false, 'unit does not exist'
	elseif not UnitIsPlayer(unit)  then
		return false, 'unit is not a player'
	end
	return self:AcceptPlayer(UnitLevel(unit), UnitFactionGroup(unit), select(2, UnitName(unit)))
end

--------------------------------------------------------------------------------
-- Score coloring and formatting
--------------------------------------------------------------------------------

local TIERS, TOP_SCORE
local COLORS = {
	{ 0.55, 0.55, 0.55 },
	{ 1.00, 1.00, 1.00 },
	{ 0.12, 1.00, 0.00 },
	{ 0.00, 0.50, 1.00 },
	{	0.69, 0.28, 0.97 },
	{ 0.94, 0.47, 0.00 },
	--{ 1.00, 0.00, 0.00 },
}
local TOP_SCORE

local function GetColor(score)
	if not addon.db.profile.qualityColor then return end
	local tiers = TIERS.PvE
	local i0, i1, f
	local numTiers = #tiers
	if score <= tiers.MIN then
		i1, f = tiers[1][2], score / tiers.MIN
		i0 = i1 - 1
	elseif score >= tiers.MAX then
		i1, f = tiers[numTiers][2], 1
		i0 = i1 - 1
	else
		for j = 2, numTiers do
			local lower, upper = tiers[j-1][1], tiers[j][1]
			if score >= lower and score < upper then
				i0, i1 = tiers[j-1][2], tiers[j][2]
				f = (score - lower) / (upper - lower)
				break
			end
		end
	end
	local r1, g1, b1 = unpack(COLORS[i0])
	local r2, g2, b2 = unpack(COLORS[i1])
	local r, g, b = r1 + f * (r2 - r1), g1 + f * (g2 - g1), b1 + f * (b2 - b1)
	return r, g, b
end

local function Colorize(text, r, g, b)
	if not text or text:trim() == "" or not r then return text end
	if type(r) == "table" then
		if type(r.r) == "number" and type(r.g) == "number" and type(r.b) == "number" then
			r, g, b = r.r, r.g, r.b
		else
			r, g, b = unpack(r, 1, 3)
		end
	end
	if type(r) == "number" and type(g) == "number" and type(b) == "number" then
		return ("|cff%02x%02x%02x%s|r"):format(255*r, 255*g, 255*b, tostring(text))
	else
		return text
	end
end
addon.Colorize = Colorize

local function ClassColorizedName(nameOrRecord, class)
	local name
	if type(nameOrRecord) == "table" then
		name, class = nameOrRecord.fullName, nameOrRecord.class
	else
		name = nameOrRecord
	end
	return Colorize(name, (CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS)[class])
end
addon.ClassColorizedName = ClassColorizedName

function addon:IsPvP(record)
	return record and (record.resilience or 0) > record.level * 2
end

function addon:InitializeTiers()
	local S = self.GetScoreForFullStuff
	TIERS = {
		PvE = {
			-- Score, color index, long label, short label
			{ S(272, ITEM_QUALITY_RARE), 2, L["Pre-Cataclysm stuff"], L["PreCat"] },
			{ S(325, ITEM_QUALITY_RARE), 3, L["Leveling stuff"], L["Lvl"] },
			{ S(333, ITEM_QUALITY_RARE), 4, L["Dungeon blues"], L["D85"] },
			{ S(346, ITEM_QUALITY_RARE), 4, L["Heroic blues"], L['HD85'] },
			{ S(359, ITEM_QUALITY_EPIC), 5, L["Tier-11 epics"], L['T11'] },
			{ S(372, ITEM_QUALITY_EPIC), 6, L["Heroic tier-11 epics"], L['HT11'] },
		},
		PvP = {
			-- Score, color index, long label, short label
			{ S(272, ITEM_QUALITY_RARE), 2, L["Pre-Cataclysm stuff"], L["PreCat"] },
			{ S(325, ITEM_QUALITY_RARE), 3, L["Leveling stuff"], L["Lvl"] },
			{ S(339, ITEM_QUALITY_RARE), 4, L["Crafted blues"] },
			{ S(352, ITEM_QUALITY_RARE), 4, L["Bloodthirsty"] },
			{ S(365, ITEM_QUALITY_EPIC), 5, L["Vicious"] },
		}
	}

	-- Calculate average values
	for kind, tiers in pairs(TIERS) do
		for i = #tiers, 2, -1 do
			tiers[i][5] = (tiers[i][1] + tiers[i-1][1]) / 2
		end
		tiers.MIN = tiers[1][1]
		tiers.MAX = tiers[#tiers][1]
	end
	
	self.ratingScheme = self:GetRatingSchemeID()

	--[===[@debug@
	self.TIERS = TIERS
	--@end-debug@]===]
end

local function EstimateTier(score, isPvP, shortFormat)
	if not score then return "?" end
	local tiers = TIERS[isPvP and "PvP" or "PvE"]
	local maxNum = #tiers
	for i = 2, maxNum do
		if score < tiers[i][5] then
			return shortFormat and tiers[i-1][4] or tiers[i-1][3]
		end
	end
	return shortFormat and tiers[maxNum][4] or tiers[maxNum][3]
end

do
	local t = {}
	function addon:FormatScore(record, noAgeSuffix, shortFormat, veryShortFormat)
		if not record or not record.score or record.score == 0 then return "?", nil, nil, nil, "?", false end
		shortFormat = shortFormat or veryShortFormat
		local fmt = self.db.profile.scoreFormat
		local isPvP = self:IsPvP(record)
		local withScore, withTier = strmatch(fmt, "score"), not veryShortFormat and strmatch(fmt, "tier")
		wipe(t)
		if withScore then
			local alternative = (record.ratingScheme ~= addon.RATING_LEGACY_QUALITY_FLAGS) and "*" or ""
			if isPvP and not withTier then
				tinsert(t, format((shortFormat and L["%dP"] or L["%d (PvP)"]), record.score)..alternative)
			else
				tinsert(t, tostring(record.score)..alternative)
			end
		end
		if withTier then
			tinsert(t, EstimateTier(record.score, isPvP, shortFormat))
		end
		if not veryShortFormat and strmatch(fmt, "itemlevel") then
			if isPvP and not withScore and not withTier then
				tinsert(t, format((shortFormat and L["%dP"] or L["%d (PvP)"]), record.average))
			else
				tinsert(t, format("%d", record.average))
			end
		end
		local scoreText = table.concat(t, " / ")
		local age = time() - record.timestamp
		local ageSuffix = (age > 0) and format(L["%s ago"], SecondsToTime(age, false, false, 1)) or L["now"]
		if not shortFormat and not noAgeSuffix and self.db.profile.showDate then
			scoreText = strjoin("", scoreText, " (", ageSuffix, ")")
		end
		local r, g, b = GetColor(record.score)
		return scoreText, r, g, b, ageSuffix
	end
end

function addon:ReportToChat(record)
	if not record then return end
	local name = ClassColorizedName(("%s (%d)"):format(record.fullName, record.level), record.class)
	self:Feedback(name, L["Gear score: "], Colorize(self:FormatScore(record)))
end

function addon:ChatQuery(input)
	if not input or input:trim() == "" then return end
	local fullName, name, realm = NormalizeName(input)
	if name:match('%*') or realm:match('%*') then
		local count = 0
		for name, realm, record in self:SearchDatabase(name, realm) do
			count = count + 1
			if count > 20 then
				break
			end
			self:ReportToChat(record)
		end
		self:WipeCache()
		if count > 20 then
			return self:Feedback(L["More than 20 matchs, please refine your query."])
		elseif count > 0 then
			return self:Feedback(L["%d matchs."]:format(count))
		end
	else
		local record = self:QueryRecord(name, realm)
		if record then
			return self:ReportToChat(record)
		end
	end
	self:Feedback('No character matched your query: '..input)
end

--------------------------------------------------------------------------------
-- Centralized character data querying
--------------------------------------------------------------------------------

local pendingQueries = {}
local function IsPendingQuery(fullName)
	return (pendingQueries[fullName] or 0) >= GetTime() - 20
end

function addon:QueryRecord(name, realm, unit)
	local fullName
	if not name or name == "" then
		if not unit then return end
		name, realm, fullName = UnitNameRealm(unit)
		if not name then return end
	else
		fullName = NormalizeName(name, realm)
	end
	local record = self:GetPlayerRecord(name, realm)
	local pending = IsPendingQuery(fullName)
	if record then
		pendingQueries[fullName] = nil
		if pending and self.db.profile.chatFeedback then
			self:ReportToChat(record)
		end
		return record
	elseif not pending then
		self:Debug('Querying record for', fullName, unit)
		pendingQueries[fullName] = GetTime()
		self:SendMessage('YAGRA_QueryRecord', name, realm, unit)
	end
end

function addon:YAGRA_RecordUpdated(event, record, source)
	local fullName = NormalizeName(record.name, record.realm)
	if IsPendingQuery(fullName) and self.db.profile.chatFeedback then
		self:ReportToChat(record)
	end
	pendingQueries[fullName] = nil
end

--------------------------------------------------------------------------------
-- Module setup
--------------------------------------------------------------------------------

local modulePrototype = {
	Debug = addon.Debug,
	Feedback = addon.Feedback,
	RegisterDatabase = function(self, defaults)
		self.db = addon.db:RegisterNamespace(self.name, defaults)
		return self.db
	end,
	IsConfigEnabled = function(self)
		return addon.db.profile.enabled and addon.db.profile.modules[self.moduleName]
	end,
}
modulePrototype.prototype = modulePrototype

addon.modulePrototype = modulePrototype
addon:SetDefaultModulePrototype(modulePrototype)
addon:SetDefaultModuleLibraries('AceEvent-3.0')

function addon:OnModuleCreated(module)
	self:Debug('OnModuleCreated', module.moduleName)
	self:RegisterModuleOptions(module)
end

-- Enable/disable module depending on combat lockdown and settings
function addon:UpdateModuleStates(event)
	local inCombat = (event == 'PLAYER_REGEN_DISABLED') or InCombatLockdown()
	local onInitialize = (event == 'OnInitialize')
	local enabledModules = self.db.profile.modules
	local addonEnabled = self:IsEnabled()
	for name, module in self:IterateModules() do
		local enableModule = enabledModules[name] and (not inCombat or module.enabledInCombat)
		if not addonEnabled or onInitialize then
			module:SetEnabledState(enableModule)
		elseif enableModule then
			module:Enable()
		else
			module:Disable()
		end
	end
end
