--[[
YAGRA - gear rating addon
Copyright 2010 Adirelle (adirelle@tagada-team.net)
All rights reserved.
--]]

local QUERY_DELAY = 3

local addonName, addon = ...
local L = addon.L

local mod = addon:NewModule('Communication', 'AceTimer-3.0')
mod.enabledInCombat = true

local ChatThrottleLib = ChatThrottleLib

local playerName, playerFaction, playerRealm = addon.playerName, addon.playerFaction, addon.playerRealm

function mod:OnInitialize()
	self:RegisterDatabase{
		profile = {
			push = true,
			query = true,
		}
	}
end

function mod:OnEnable()
	self:RegisterEvent('CHAT_MSG_ADDON')
	self:RegisterMessage('YAGRA_QueryRecord')
	self:RegisterMessage('YAGRA_RecordUpdated')
	self:RegisterEvent('PLAYER_ENTERING_WORLD', 'UpdateBroadcast')
	self:RegisterEvent('PARTY_MEMBERS_CHANGED', 'UpdateBroadcast')
	self:UpdateBroadcast()
end

local function GetBroadcastChannel()
	local _, instanceType = IsInInstance()
	if instanceType == "pvp" then
		return "BATTLEGROUND"
	elseif GetRealNumRaidMembers() > 0 then
		return "RAID"
	elseif instanceType == "arena"  or GetRealNumPartyMembers() > 0 then
		return "PARTY"
	elseif IsInGuild() then
		return "GUILD"
	end
end

function mod:UpdateBroadcast()
	local broadcast = GetBroadcastChannel()
	if broadcast ~= self.broadcast then
		self.broadcast = broadcast
		if broadcast and self.db.profile.push then
			local record = addon:GetPlayerRecord(playerName, playerRealm)
			if record then
				self:SendRecord(record, broadcast)
			end
		end
	end
end

function mod:GetOptions()
	return {
		name = L['Communication'],
		type = 'group',
		order = 200,
		args = {
			enable = self:CreateEnableOption{
				name = L['Enable'],
				desc = L['Enable communications with other players to share your database.'],
				width = 'double',
				order = 201,
			},
			push = {
				name = L['Push updates'],
				desc = L['Automatically send player entry updates to raid or guild addon channel.'],
				type = 'toggle',
				order = 210,
			},
			query = {
				name = L['Query missing entries'],
				desc = L['Send query to raid and guild for entries that are not in your database.'],
				type = 'toggle',
				order = 220,
			},
		},
	}, self:GetOptionHandler()
end

local function CanSafelySend(channel, target, unit)
	if (channel == "GUILD" and IsInGuild()) or
			((channel == "RAID" or channel == "PARTY") and GetRealNumPartyMembers() > 0) or
			(channel == "WHISPER" and addon:IsOnline(target, unit)) then
		return true
	elseif channel == "BATTLEGROUND" then
		local _, instanceType = IsInInstance()
		if instanceType == "pvp" or instanceType == "arena" then
			return true
		end
	end
end

local function SendAddonMessage(prefix, message, channel, target)
	mod:Debug('SendAddonMessage', prefix, tostring(message):sub(1,30), channel, target)
	ChatThrottleLib:SendAddonMessage("BULK", prefix, message, channel, target, mod.name)
end

local tstruct = {}
local function ParseDate(dateString)
	local year, month, day, hour, min = tostring(dateString):match('^(%d%d%d%d)(%d%d)(%d%d)(%d%d)(%d%d)$')
	if not tonumber(year) then return end
	tstruct.year = tonumber(year)
	tstruct.month = tonumber(month)
	tstruct.day = tonumber(day)
	tstruct.hour = tonumber(hour)
	tstruct.min = tonumber(min)
	tstruct.sec = 0
	local timestamp = time(tstruct)
	return timestamp
end

function mod:InsertRecord(realm, name, score, dateString, class, average, race, faction, location, level, guild, source, ...)
	if name == playerName and realm == playerRealm then
		return 'player', name
	end
	local timestamp = ParseDate(dateString)
	local record = addon:GetPlayerRecord(name, realm)
	local new = not record
	if new then
		record = {}
	elseif record.timestamp and record.timestamp >= timestamp then
		return 'date', name
	end
	record.score = tonumber(score)
	record.timestamp = timestamp
	record.class = addon.UNSERIALIZED_CLASS[class]
	record.average = tonumber(average)
	record.race = addon.UNSERIALIZED_RACE[race]
	record.faction = addon.UNSERIALIZED_FACTION[faction]
	record.level = tonumber(level)
	for i = 1, 18 do
		record[i] = select(i, ...) or "0:0"
	end
	if new then
		if addon:IsRecordValid(record) then
			 addon:CreatePlayerRecord(name, realm, record)
		else
			return 'invalid', name
		end
	end
	return false, name, record
end

local function FormatDate(timestamp)
	local d = date("*t", timestamp)
	return ("%04d%02d%02d%02d%02d"):format(d.year, d.month, d.day, d.hour, d.min)
end

local function SerializeItems(item, ...)
	item = item or "0:0"
	if select('#', ...) > 0 then
		return item, SerializeItems(...)
	else
		return item
	end
end

function mod:SendRecord(record, channel, target)
	local message = strjoin("$",
		tostringall(
			record.name,
			record.score,
			FormatDate(record.timestamp),
			addon.SERIALIZED_CLASS[record.class],
			record.average,
			addon.SERIALIZED_RACE[record.race],
			addon.SERIALIZED_FACTION[record.faction],
			'XXX',
			record.level,
			'*',
			playerName,
			SerializeItems(unpack(record, 1, 18))
		)
	)
	SendAddonMessage("GSY", message, channel, target)
end

local pendingQueries = {}

function mod:SendQuery(name)
	self:Debug('Sending query for', name)
	pendingQueries[name] = nil
	if self.broadcast then
		SendAddonMessage("GSY_Request", name, self.broadcast)
	end
end

function mod:YAGRA_QueryRecord(event, name, realm, unit)
	if self.db.profile.query and self.broadcast and realm == playerRealm and not pendingQueries[name] then
		self:Debug('Postponing query for', name)
		pendingQueries[name] = self:ScheduleTimer('SendQuery', QUERY_DELAY, name)
	end
end

function mod:YAGRA_RecordUpdated(event, record, source)
	if record.realm == playerRealm then
		self:CancelTimer(pendingQueries[record.name], true)
		pendingQueries[record.name] = nil
		if source ~= self.name and self.db.profile.push and self.broadcast and not addon.db.profile.alternativeRating then
			self:Debug('Sending updated record of', record.name)
			self:SendRecord(record, self.broadcast)
			if self.broadcast ~= "GUILD" and IsInGuild() then
				self:SendRecord(record, "GUILD")
			end
		end
	end
end

function mod:CHAT_MSG_ADDON(event, prefix, message, channel, sender)
	-- Do not listen to ourselves nor players from other servers
	if sender == playerName or sender:match('-') then return end
	local handler = self["OnReceive_"..prefix]
	if handler then
		self:Debug('CHAT_MSG_ADDON', prefix, tostring(message):sub(1, 30), channel, sender)
		return handler(self, message, channel, sender)
	end
end

function mod:OnReceive_GSY_Request(name, channel, sender)
	local record = addon:GetPlayerRecord(name, playerRealm)
	if record then
		self:Debug('Sending', name, 'record to', sender)
		self:SendRecord(record, channel, sender)
	end
end

function mod:OnReceive_GSY(message, channel, sender)
	local failure, name, record = self:InsertRecord(playerRealm, strsplit('$', message))
	if not failure then
		self:Debug('Accepted received', name, 'record from', sender)
		addon:UpdatePlayerScore(record, self.name)
	else
		self:Debug('Ignored received', name, 'record from', sender, 'because of', failure)
	end
end
