--[[
YAGRA - gear rating addon
Copyright 2010 Adirelle (adirelle@tagada-team.net)
All rights reserved.
--]]

local addonName, addon = ...
local L = addon.L

-- Record time-to-live, i.e. time before a character record is considered old enough to be refreshed
local RECORD_TTL = 600

-- Scanning queue processing period, in seconds
local QUEUE_PROCESSING_PERIOD = 2

-- Minimum time between latest INSPECT_READY event and NotifyInspect, in seconds
local INSPECT_COOLDOWN = 5

local UnitNameRealm = addon.UnitNameRealm

local mod = addon:NewModule('Scanner', 'AceTimer-3.0', 'AceBucket-3.0', 'AceHook-3.0')

local ENCHANTABLE = {
	INVTYPE_HEAD = true,
	INVTYPE_SHOULDER = true,
	INVTYPE_CHEST = true,
	INVTYPE_ROBE = true,
	INVTYPE_LEGS = true,
	INVTYPE_FEET = true,
	INVTYPE_WRIST = true,
	INVTYPE_HAND = true,
	INVTYPE_CLOAK = true,
	INVTYPE_WEAPON = true,
	INVTYPE_2HWEAPON = true,
	INVTYPE_WEAPONMAINHAND = true,
	INVTYPE_WEAPONOFFHAND = true,
	INVTYPE_SHIELD = true,
	INVTYPE_RANGED = true,
}

local inspectQueue, failedQueue = {}, {}

function mod:OnInitialize()
	self:RegisterDatabase{
		profile = {
			group = true,
		}
	}
end

function mod:OnEnable()
	self:RegisterBucketEvent('PARTY_MEMBERS_CHANGED', 5)
	self:RegisterBucketEvent('UNIT_INVENTORY_CHANGED', 0.5)
	self:RegisterEvent('INSPECT_READY')
	self:RegisterMessage('YAGRA_QueryRecord')
	self:RegisterMessage('YAGRA_RecordUpdated')
	self:ScheduleRepeatingTimer('ProcessQueue', QUEUE_PROCESSING_PERIOD)
	self:QueueUnit('player')
end

function mod:OnDisable()
	wipe(inspectQueue)
	wipe(failedQueue)
end

function mod:GetOptions()
	return {
		_scannerHeader = {
			name = L['Scanning'],
			type = 'header',
			order = 100,
		},
		enableScanning = self:CreateEnableOption{
			name = L['Enable scanning'],
			desc = L['Enable scanning of other players either on demand or automatically.'],
			order = 110,
		},
		scanGroup = {
			name = L['Scan group'],
			desc = L['Actively scan players in your group.'],
			type = 'toggle',
			arg = 'group',
			order = 120,
		},
	}, self:GetOptionHandler()
end

function mod:UNIT_INVENTORY_CHANGED(units)
	self:Debug(event)
	if units.player then
		self:Inspect("player", UnitGUID("player"))
	end
end

local function CanInspectUnit(unit)
	mod:Debug("CanInspectUnit", unit, "exists=", UnitExists(unit), "player=", UnitIsPlayer(unit), "visible=", UnitIsVisible(unit), "connected=", UnitIsConnected(unit), "distance=", CheckInteractDistance(unit, 4))
	return UnitExists(unit) and UnitIsPlayer(unit) and UnitIsVisible(unit) and UnitIsConnected(unit) and CheckInteractDistance(unit, 4)
end

local lastInspectedGUID
local items = {}
function mod:Inspect(unit, guid)
	if not unit or not guid then return end
	inspectQueue[guid], failedQueue[guid] = nil, nil
	if not self:ScanInventory(unit, items) then
		self:Debug('[Scan] Inspect data are incomplete', unit)
		-- Put the unit back in queue
		failedQueue[guid] = true
		return
	end
	lastInspectedGUID = guid

	local name, realm = UnitNameRealm(unit)
	local timestamp = time()
	local record = addon:GetPlayerRecord(name, realm)
	if not record then
		record = addon:CreatePlayerRecord(name, realm)
		self:Debug('[Scan] Creating new record for', name, realm)
	elseif record.timestamp >= timestamp then
		self:Debug('[Scan] Record is up to date for', name, realm)
		return
	--[===[@debug@--
	else
		self:Debug('[Scan] Updating record of', name, realm)
	--@end-debug@]===]--
	end
	local level = UnitLevel(unit)
	local updated = (record.level ~= level)
	record.level = level
	record.faction = UnitFactionGroup(unit)
	record.class = select(2, UnitClass(unit))
	record.race = select(2, UnitRace(unit))
	record.timestamp = timestamp
	for slot = 1, 18 do
		local item = items[slot]
		if record[slot] ~= item then
			record[slot] = item
			updated = true
		end
	end
	if updated or not record.score or not record.average then
		return addon:UpdatePlayerScore(record, self.name)
	end
	self:Debug('[Scan] No change for', record.name)
end

local function GetUnitForGUID(guid, noMouseover)
	if UnitGUID("player") == guid then
		return "player"
	end
	local prefix, max = "raid", GetNumRaidMembers()
	if max == 0 then
		prefix, max = "party", GetNumPartyMembers()
	end
	if max > 0 then
		for i = 1, max do
			if UnitGUID(prefix..i) == guid then
				return prefix..i
			end
		end
	end
	if UnitGUID("target") == guid then
		return "target"
	elseif UnitGUID("focus") == guid then
		return "focus"
	elseif not noMouseover and UnitGUID("mouseover") == guid then
		return "mouseover"
	end
end

local lastInspectTime = 0
function mod:INSPECT_READY(event, guid)
	lastInspectTime = time()
	local unit = GetUnitForGUID(guid)
	if not unit then
		self:Debug('INSPECT_READY ignored: not unit for guid', guid)
		return
	end
	local valid, reason = addon:IsValidUnit(unit)
	if not valid then
		self:Debug('INSPECT_READY ignored for', guid, ':', reason)
		return
	end
	self:Debug('INSPECT_READY', guid, unit)
	self:Inspect(unit, guid)
end

function mod:ScanInventory(unit, items)
	local incomplete, missing = 0, 0
	for slot = 1, 18 do
		items[slot] = "0:0"
		if slot ~= INVSLOT_BODY then
			if GetInventoryItemID(unit, slot) then
				local link = GetInventoryItemLink(unit, slot)
				if link and not strmatch(link, "item:%d+:0:0:0:0:0:0:0:0:0") then
					items[slot] = strmatch(link, 'item:(%d+:%d+)')
				else
					self:Debug('[Scan] Missing or incomplete item link:', slot, link)
					incomplete = incomplete + 1
				end
			elseif slot ~= INVSLOT_OFFHAND then
				self:Debug('[Scan] Empty slot:', slot)
				missing = missing + 1
			end
		end
	end
	if items[INVSLOT_MAINHAND] and not items[INVSLOT_OFFHAND] and select(9, GetItemInfo(items[INVSLOT_MAINHAND])) ~= "INVTYPE_2HWEAPON" then
		missing = missing + 1
		self:Debug('[Scan] Missing off-hand')
	end
	self:Debug('[Scan] Result:', missing, 'missing item(s)', incomplete, 'incomplete link(s)')
	if incomplete > 0 or missing > 0 then
		wipe(items)
		return false
	else
		return true
	end
end

function mod:QueueUnit(unit)
	local name, realm, fullName = UnitNameRealm(unit)
	local valid, reason = addon:IsValidUnit(unit)
	if not valid then
		return self:Debug('[Queue] Ignoring invalid unit', unit, fullName, reason)
	end
	local guid = UnitGUID(unit)
	if not guid then
		return self:Debug('[Queue] Unit has not GUID!', unit, fullName)
	end
	if unit == 'mouseover' then
		unit = GetUnitForGUID(guid, true)
		if not unit then
			self:Debug('[Queue] Not scanning mouseover')
			return
		end
	end
	if not inspectQueue[guid] and not failedQueue[guid] then
		local record = addon:GetPlayerRecord(name, realm)
		if not record or (record.score or 0) == 0 or time() - record.timestamp > RECORD_TTL then
			inspectQueue[guid] = true
		--[===[@debug@
			self:Debug('[Queue] Queued for inspection:', unit, fullName)
		else
			self:Debug('[Queue] Existing record is fresh enough:', unit, fullName)
		--@end-debug@]===]
		end
	end
end

local function ShiftQueue()
	local switched = false
	local guid, unit
	repeat
		guid = next(inspectQueue)
		if not guid and not switched then
			guid = next(failedQueue)
			if guid then
				mod:Debug('[Queue] trying postponed units again')
				switched = true
				inspectQueue, failedQueue = failedQueue, inspectQueue
			end
		end
		if not guid then
			return
		end
		inspectQueue[guid] = nil
		unit = GetUnitForGUID(guid)
		if not unit then
			mod:Debug('[Queue] skipped', guid,': cannot find an unit')
		elseif not addon:IsValidUnit(unit) then
			mod:Debug('[Queue] skipped', unit,': invalid')
			guid = nil
		elseif not CanInspectUnit(unit) then
			mod:Debug('[Queue] postponing', unit,': cannot inspect right now')
			failedQueue[guid], guid = true, nil
		elseif guid == lastInspectedGUID then
			mod:Debug('[Queue] skipped', unit,': latest unit scanned')
			failedQueue[guid], guid = true, nil
		end
	until unit and guid
	mod:Debug('[Queue] Ready to inspect:', unit, guid, pcall(UnitNameRealm, unit))
	return unit, guid
end

function mod:ProcessQueue()
	if not InCombatLockdown() and not (InspectFrame and InspectFrame:IsVisible()) then
		if time() - lastInspectTime > INSPECT_COOLDOWN then
			local unit, guid = ShiftQueue()
			if unit == 'player' and guid then
				self:Inspect(unit, guid)
			elseif unit then
				self:Debug('[Queue] NotifyInspect', unit)
				NotifyInspect(unit)
			end
		end
	else
		self:Debug('[Queue] Paused')
	end
end

function mod:PARTY_MEMBERS_CHANGED()
	if not self.db.profile.group then return end
	local prefix, num = 'raid', GetNumRaidMembers()
	if num == 0 then
		prefix, num = 'party', GetNumPartyMembers()
		if num == 0 then
			return
		end
	end
	for i = 1, num do
		self:QueueUnit(prefix..i)
	end
end

function mod:YAGRA_QueryRecord(event, name, realm, unit)
	if unit then
		self:QueueUnit(unit)
	end
end

function mod:YAGRA_RecordUpdated(event, record, source)
	local guid = record.guid
	if guid then
		inspectQueue[guid], failedQueue[guid] = nil
	end
end
